/*
 * CSkyV2.cpp
 *
 *  Created on: Feb 25, 2009
 *      Author: qwan
 */

#include "CSkyV2.h"

#define _MAXIMUM 65535

unsigned long csky::skyline_count = 0;
unsigned long csky::pruned = 0;

void csky::addto_vector(int* v, int tabNo, int ptNo) {
	for (int i = 0; i < tables[tabNo]->dimNum; i++) {
		v[tables[tabNo]->f[i]] += tables[tabNo]->v[ptNo][i];
	}
}
;

Pt_Collection* csky::getWorstRegion() {

	int* q = new int[tables[0]->dimNum];
	int* zero = new int[tables[0]->dimNum];

	memset(zero, 0, sizeof(int) * tables[0]->dimNum);
	memset(q, 0, sizeof(int) * tables[0]->dimNum);

	for (int i = 1; i < tableNum; i++) {
		for (int j = 0; j < tables[i]->dimNum; j++) {
			int d = tables[i]->f[j];
			q[d] += tables[i]->M[j];
		}
	}

	Pt_Collection* re = tables[0]->rect_query(zero, q);

	delete[] zero;
	delete[] q;
	//	delete[] query;

	return re;

}

void csky::preprocessing() {

	wregion = getWorstRegion();

	int size = wregion->size();

	upper = size;

	//upper = (int)floor(avg * 2) + 3;
	//lower = (int)floor(avg / 10) + 1;

	//avg = avg * 0.75 + size*0.25;


	if (size == 0) {
		type = 0;

		int full = 1;
		for (int i = 1; i < tableNum; i++) {
			full *= tables[i]->tupleNum;
		}
		csky::skyline_count += full;

		int* dl = new int[tableNum - 1];
		for (int i = 0; i < tableNum - 1; i++) {
			dl[i] = tables[i + 1]->tupleNum;
		}

		cs->init(tableNum - 1, dl);

		return;
	} else if (size < lower) {
		type = 1;
		int* dl = new int[tableNum - 1];
		for (int i = 0; i < tableNum - 1; i++) {
			dl[i] = tables[i + 1]->tupleNum;
		}

		cs->init(tableNum - 1, dl);

		return;
	} else if (size > upper) {
		type = 2;

		PtsSpace* pts = tables[0]->Product(wregion);

		int tup = pts->tupleNum;
		pts->tupleNum = upper;

		//printf("using %d instead of %d\n",pts->tupleNum, tables[0]->tupleNum);

		int* dl = new int[tableNum - 1];
		for (int i = 0; i < tableNum - 1; i++) {
			dl[i] = tables[i + 1]->tupleNum;
			int* map = tables[i + 1]->f;
			int dim = tables[i + 1]->dimNum;
			int non = tables[i + 1]->nonMerg;
			pack[i] = pts->Selection(map, dim, non);
			package.push_back(pack[i]);
		}

		pts->tupleNum = tup;
		delete pts;

		cs->init(tableNum - 1, dl);

		return;

	} else {
		type = 3;

		PtsSpace* pts = tables[0]->Product(wregion);

		//printf("using %d instead of %d\n",pts->tupleNum, tables[0]->tupleNum);

		int* dl = new int[tableNum - 1];
		for (int i = 0; i < tableNum - 1; i++) {
			dl[i] = tables[i + 1]->tupleNum;
			int* map = tables[i + 1]->f;
			int dim = tables[i + 1]->dimNum;
			int non = tables[i + 1]->nonMerg;
			pack[i] = pts->Selection(map, dim, non);
			package.push_back(pack[i]);
		}

		delete pts;

		cs->init(tableNum - 1, dl);

		return;

	}

}

void csky::run() {

	preprocessing();

	if (type == 0) { //match when group size is small

		int* buf;

		while (!cs->isoverflow()) {
			int* tp = cs->next();

			buf = new int[tables[0]->dimNum];
			memset(buf, 0, sizeof(int) * tables[0]->dimNum);

			for (int j = 0; j < cs->dim; j++) {
				addto_vector(buf, j + 1, tp[j]);
			}

			//result.push_back(buf);


			delete[] buf;
			delete[] tp;

		}

		return;
	} else if (type == 1) { //never matched!!!!!!!!!!!

		int* buf = new int[tables[0]->dimNum];

		while (!cs->isoverflow()) {
			int* tp = cs->next();
			memset(buf, 0, sizeof(int) * tables[0]->dimNum);
			if (tp != NULL) {

				for (int j = 0; j < cs->dim; j++) {
					addto_vector(buf, j + 1, tp[j]);
				}

				csky::pruned++;
				if (tables[0]->rect_checking(buf)) {
					csky::skyline_count++;
				} else {
					//				for(int i=0;i<3;i++){
					//					printf("%d ",tp[i]);
					//				}
					//				for(int i=0;i<tables[0]->dimNum;i++){
					//					printf("%d ",buf[i]);
					//				}
				}

				//printf("\n");

				delete[] tp;
			}
		}
		delete[] buf;

		return;

	} else {
		Pt_Collection** array = new Pt_Collection*[tableNum - 1];
		bool insertNeeded = true;
		int* max = new int[20];
		int* buf = new int[tableNum - 1];

		for (int i = 0; i < 20; i++) {
			max[i] = _MAXIMUM;
		}

		int tup = package[0]->tupleNum;

		for (int i = 0; i < tup; i++) {
			for (int j = 0; j < tableNum - 1; j++) {

				int* test = package[j]->v[i];

				array[j] = tables[j + 1]->rect_query(test, max);
				if (array[j]->size() == 0) {
					insertNeeded = false;
					for (int k = j; k >= 0; k--) {
						delete array[k];
					}

					break;
				} else {
					//printf("i=%d,j=%d\n",i,j);
				}
			}

			if (insertNeeded) {
				recurInsert(&array, &buf, 0);
				for (int k = 0; k < tableNum - 1; k++) {
					delete array[k];
				}
			}

			insertNeeded = true;

		}

		delete[] array;
		delete[] max;
		delete[] buf;

		buf = new int[tables[0]->dimNum];

		while (!cs->isoverflow()) {
			int* tp = cs->next();
			if (tp != NULL) {

				memset(buf, 0, sizeof(int) * tables[0]->dimNum);

				for (int j = 0; j < cs->dim; j++) {
					addto_vector(buf, j + 1, tp[j]);
				}

				csky::pruned++;
				if (tables[0]->rect_checking(buf)) {
					//result.push_back(buf);
					csky::skyline_count++;
				} else {

				}

				delete[] tp;
			}
		}

		delete[] buf;

	}

}

void csky::recurInsert(Pt_Collection*** parray, int** pbuf, int depth) {

	Pt_Collection** array = *parray;
	int* buf = *pbuf;

	if (depth == tableNum - 1) {
		int index = cs->mapping(buf);
		cs->setBit(index, true);
	} else {
		Pt_Collection::iterator iter;
		for (iter = array[depth]->begin(); iter != array[depth]->end(); iter++) {
			buf[tableNum - depth - 2] = *iter;
			recurInsert(parray, pbuf, depth + 1);
		}
	}

	return;
}

void csky::clear() {

	delete wregion;

	if (type == 0) {
		tables.clear();
		cs->clear();
		return;
	} else if (type == 1) {
		tables.clear();
		cs->clear();
	} else {
		//int size = package.size();
		for (int i = 0; i < tableNum - 1; i++) {
			//	if(package[i]->tupleNum != 0)
			delete pack[i];
		}
		package.clear();
		tables.clear();
		cs->clear();
	}

	return;
}

