/*
 * space.h
 *
 *  Created on: Jan 14, 2009
 *      Author: qwan
 */

#ifndef SPACE_H_
#define SPACE_H_

#include "common.h"
#include "rtree.h"

using namespace std;

//class Table contains all data about single table and its points, and the methods on those points.
//compute skyline

void free_tree(node_type *node, rtree_info *aInfo);

class PtsSpace {

public:
	int dimNum; //num of dim
	int tupleNum; //num of tuple
	int nonMerg; //num of nonMerg
	int oldDim;
	bool isPackage;
	bool isStored;
	bool indexed_flag;
	int sortpos;

	int** v; //data
	int** v2;
	int* f; //package dim related on each dim
	int* m; //minimum for each nonMerg dim;
	int* M;
	int* min;
	int** stored;
	PtsSpace* PC;
	Pt_Collection* sample;
	Pt_Collection* partition;

	int** r_data;
	node_type* r_root;
	rtree_info r_aInfo;
	stack<int*> mem;
	stack<Pt_Collection*> pool;

	PtsSpace() {
		PtsSpace::cnt++;
		init(0, 0, 0);
		sample = NULL;
	}

	PtsSpace(const PtsSpace &cp) {
		PtsSpace::cnt++;
		init(cp.dimNum, cp.tupleNum, cp.nonMerg);
		sortpos = cp.sortpos;
		for (int i = 0; i < tupleNum; i++) {
			memcpy(v[i], cp.v[i], sizeof(int) * dimNum);
		}

		memcpy(f, cp.f, sizeof(int) * dimNum);
		memcpy(m, cp.m, sizeof(int) * (dimNum - nonMerg));
		memcpy(M, cp.M, sizeof(int) * dimNum);
		memcpy(min, cp.min, sizeof(int) * dimNum);

	}

	~PtsSpace() {

		PtsSpace::cnt--;

		for (int i = 0; i < tupleNum; i++){
			delete[] v[i];
		}

		delete[] v;
		delete[] m;
		delete[] M;
		delete[] min;
		delete[] f;
		//		if(isStored){
		for (int i = 0; i < tupleNum; i++) {
			delete[] stored[i];
		}
		delete[] stored;

		release_index();

		if (PC != NULL) {
			delete PC;
		}

		if (partition != NULL) {
			delete[] partition;
		}

	}

	void getdata(FILE* fp); //get data from file
	void calMin();
	void skylinefilter(); //calculate skyline
	void prunning();
	void addto_dim(int d, int mount); //change no d dim in the table by mount
	void swap(int d1, int d2); //swap two dims
	PtsSpace* split(int n, int m, int* g); //split the table into n pieces;
	PtsSpace* split(int n); //split the table into n pieces;


	void reverseSort(int d);
	void sort(int d); //sort by d-th dim
	void calM();
	void sampling();
	double sp_dist2(int* v1, int* v2);

	PtsSpace* getSample();

	PtsSpace* Product(Pt_Collection* p); //split the table into n pieces;
	PtsSpace* ProductOrg(int index); //split the table into n pieces;
	PtsSpace* Selection(int* map, int dim, int non);
	void minusMin();

	Pt_Collection* rect_query(const int* v1, const int* v2);
	bool rect_checking(int* q);
	void build_index();
	void release_index();
	Pt_Collection* rect_query_indexed(const int* v1, const int* v2);

	void show() {
		printf("dim = %d, tup = %d, non = %d \n", dimNum, tupleNum, nonMerg);
		printf("\nv:\n");

		for (int j = 0; j < tupleNum; j++) {
			printVec(v[j], oldDim);
			printf("\n");
		}

		if (isStored) {
			for (int j = 0; j < tupleNum; j++) {
				printVec(stored[j], oldDim);
				printf("\n");
			}
		}

		printf("\nf:\n");
		printVec(f, oldDim);
		printf("\nm:\n");

		printf("\nM:\n");
		printVec(M, oldDim);
		printf("\nm:\n");

		printf("\nmin:\n");
		printVec(min, oldDim);
		printf("\nm:\n");

		printVec(m, oldDim - nonMerg);
		printf("\n");

	}

	void init(int d, int t, int n) {
		dimNum = d;
		tupleNum = t;
		nonMerg = n;
		oldDim = d;
		isPackage = false;
		isStored = false;
		indexed_flag = false;
		PC = NULL;
		partition = NULL;

		if (d * t != 0) {
			v = new int*[t];
			stored = new int*[t];

			for (int i = 0; i < t; i++) {
				v[i] = new int[d];
				stored[i] = new int[d];
			}

			f = new int[d];
			m = new int[d - n];
			M = new int[d];
			min = new int[d];

		}

	}

	void tpinit(int d, int t, int n) {
		dimNum = d;
		tupleNum = t;
		nonMerg = n;
		oldDim = d;
		isPackage = false;
		isStored = false;
		indexed_flag = false;
		PC = NULL;
		partition = NULL;

		if (d * t != 0) {
			v = new int*[t];

			//		for(int i=0;i<t;i++){
			//		  v[i]=new int[d];
			//		}

		}

	}

	void store();
	void restore();
	long count() {
		long size = 0;
		size += sizeof(int) * tupleNum * oldDim * 2;
		size += sizeof(int) * oldDim * 3;
		size += sizeof(int) * (oldDim - nonMerg) * 2;
		size += sizeof(int*) * tupleNum * 2;
		return size;
	}

	int dominate(const int* v1, const int* v2);
	int dominate(const int* v1, const int* v2, int n);
	int dnlcomp(int* vec1, int*vec2);

	void printVec(int* v, int n);

	void rectangle_search(node_type*curr_node, const int* u, const int* l,
			Pt_Collection* result, bool inside);

	void addpoint(int* len, int ptr);
	void bnl();
	double calDiff();
	double calDiff2();
	void remove(Pt_Collection& s);
	bool rectangle_find(node_type*curr_node, const int* u, const int* l);

	void tofile(const char* fileName) {
		FILE* f = fopen(fileName, "w");

		fprintf(f, "%d %d\n", tupleNum, dimNum);

		for (int i = 0; i < tupleNum; i++) {
			for (int j = 0; j < dimNum; j++) {
				fprintf(f, "%d ", v[i][j]);
			}
			fprintf(f, "\n");
		}
		fclose(f);
	}
	;

	Pt_Collection* KMeans(int n);
	Pt_Collection* KMeans2(int n);

	double sp_dist(int* v1, int* v2);

	void kmp() {
		partition = KMeans(4);

		printf("\npackage partition done!!!\n");
	}

	static unsigned long cacheHits;
	static unsigned long totalAsked;
	static unsigned long cacheMissed;

	static int cnt;
};

#endif /* SPACE_H_ */
